/******************************************************************************
 *    File: prgrun.c                                 an addition to CircleMUD *
 *   Usage: Asynchronyously run unix programs from MUD                        *
 *  Author: Petr Vilim (Petr.Vilim@st.mff.cuni.cz)                            *
 *                                                                            *
 *****************************************************************************/

/*
 * WARNING 1: It works on Linux. It should work on all UNIXes but 
 *            I didn't try it.
 *
 * WARNING 2: Use on your own risk :-)
 *
 * WARNING 3: Playing with this you can easily make some security hole, 
 *            if you make some command modifying spefied file for example. 
 *            Mud password travels through net uncrypted and can be easily 
 *            caught (try tcpdump). So some hacker can log on mud with your 
 *            char and can destroy all your files...  (and can modify 
 *            file ~/.klogin and I don't know what more) 
 *            This is the reason why I didn't make command exec which 
 *            calls function system with given parametrs.
 */
 
 /*
  * You haven't to give me a credit. If you want you can mail me.
  */

// Adapted to CircleMUD-3.1 - 07/03 - (corporeal@impmud)
//
// This is actually the stock prgrun.c slightly modified here
// and there for the 3.1 port.  Our (impmud's) version of this
// is so hopelessly different that I gave up on just porting 
// our version.  
//
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "conf.h"
#include "sysdep.h"
#include "structs.h"
#include "utils.h"
#include "comm.h"
#include "interpreter.h"
#include "db.h"
#include "oasis.h"
#include "genzon.h"

//#include <iostream>

#define PQUEUE_LENGTH    20  /* Max. number of requests in queue */
#define MAX_ARGS         5   /* Max. number of arguments of program */

extern struct char_data *get_ch_by_id(long idnum);

struct program_info {
  char *args[MAX_ARGS+1];/* args[0] is name of program, 
                            rest are arguments, after last argument
                            must be NULL  */
  char *input;           /* Input sended to process 		*/
  long char_id;          /* ID of character 			*/
  char *name;            /* Name used when printing output      */
  int timeout;           /* Timeout in seconds 			*/
};

struct program_info program_queue[PQUEUE_LENGTH];
int pqueue_counter=0;	 /* Number of requests in queue */

int pid=-1;		 /* PID of runnig command, -1 if none 		*/
int to[2], from[2];	 /* file descriptors of pipes between processes */
time_t start_time;	 /* Time when process started, for timeout 	*/
int had_output;		 /* Flag if running command already had an output */
FILE *process_in;        /* Standart input for runned program 		*/

/*
 * get_ch_by_id : given an ID number, searches every descriptor for a
 *              character with that number and returns a pointer to it.
 */
struct char_data *get_ch_by_id(long idnum)
{
  struct descriptor_data *d;
  extern struct descriptor_data *descriptor_list;

  for (d = descriptor_list; d; d = d->next)
    if (d && d->character && GET_IDNUM(d->character) == idnum){
      //      fprintf(stderr,"Found character\r\n.");
      return (d->character);
    }
  //  fprintf(stderr,"Did not find character\r\n");
  return NULL;
}

void program_fork() {	 /* Start new command */
  int i;
  long flags;
  //  fprintf(stderr,"Entering program_fork");
  pipe(to);
  pipe(from);
  start_time=time(0);
  had_output=0;
  pid=fork();
  if (pid < 0) {
    pid=-1;		 /* Some problem with fork */
    return;
  }
  if (pid==0) {		 /* Child process */
    dup2 (from[1], 1);	 /* Set from[1] as standart output */
    // I want to see stderr as well
    dup2 (from[1], 2);
    close (from[0]);
    close (from[1]);
    dup2 (to[0], 0);
    close (to[0]);
    close (to[1]);
    // Uh oh, no sense in using stderr if you turn around and close it!
    //    for (i=2; i<1000; i++)
    for (i=3; i<1000; i++)
      close(i);		 /* Close all opened file descriptors
			    1000 should be enough I hope :-) */
    execvp(program_queue[0].args[0], program_queue[0].args);	
    exit(0);
  }
  close(from[1]);
  close(to[0]);
  process_in=fdopen(to[1], "w");
  setbuf(process_in, NULL);
  if (program_queue[0].input)   /* Send input to program if any */
    fprintf(process_in, program_queue[0].input);
  flags=fcntl(from[0], F_GETFL);
  fcntl(from[0], F_SETFL, flags | O_NONBLOCK);
			 /* Set from[0] into no-blocking mode */
}

void program_done() {
  int i;
  
  if (pid!=-1) {
    close(from[0]);	   /* Close process standart output */
    fclose(process_in);	   /* Close process standart input  */
    waitpid(pid, NULL, 0); /* Wait for process termination  */
    for (i=0; i<MAX_ARGS+1; i++)
      if (program_queue[0].args[i]) free(program_queue[0].args[i]);
      else break;
    free(program_queue[0].name);
    if (program_queue[0].input) free(program_queue[0].input);
    for (i=0; i<pqueue_counter; i++)
      program_queue[i]=program_queue[i+1]; /* shift queue */
    pqueue_counter--;
    if (pqueue_counter)
      program_fork();	 /* Start next process */
    else
      pid=-1;  
  }
}

void process_program_output() {
  int len;
  struct char_data *ch;
  char buf[MAX_STRING_LENGTH+1],buf1[MAX_STRING_LENGTH+1];
  char *c, *d;
  //  fprintf(stderr,"Entering process_program_output");
  if (pid==-1) 
    return;
  len=read(from[0], buf, MAX_STRING_LENGTH);
  if ((len==-1) && (errno==EAGAIN)) { /* No datas are available now */
    if (time(0)<start_time+program_queue[0].timeout)
      return;
    else
      snprintf(buf, sizeof(buf),"&wKilling &c%s&w because of timeout.\r\n",
	       program_queue[0].name);
  } else if (len < 0) {	 /* Error with read or timeout */
    snprintf(buf,sizeof(buf), "&wError with reading from &c%s&w, killed.",
	    program_queue[0].name);
  } else if (len == 0) {  /* EOF readed  */
    if (had_output)
      buf[0]=0;
    else
      snprintf(buf, sizeof(buf),"&wNo output from &c%s&w.\r\n", program_queue[0].name);    
  };
  ch=get_ch_by_id(program_queue[0].char_id);
  if (len<=0) {
    kill(pid, 9);         /* kill process with signal 9 if is still running */
    program_done();
    if ((ch) && (buf[0]))
      send_to_char(ch,buf);
    return;
  }
  had_output=1;
  buf[len]='\0';
  for (c=buf, d=buf1; *c; *d++=*c++)
    if (*c=='\n') *d++='\r';
  *d=0;
  if (!ch){
    //    fprintf(stderr,"character not found");
    return;	 /* Player quited the game? */
  }
  send_to_char(ch,"&wOutput from &c");
  send_to_char(ch,program_queue[0].name);
  send_to_char(ch,"&w:&g\r\n");
  send_to_char(ch,buf1);
  send_to_char(ch,"&n");
}

/* Following function add program into queue */
void add_program(struct program_info prg, struct char_data *ch)
{
  //  fprintf(stderr,"entering add_program");
  prg.char_id=GET_IDNUM(ch);
  if (pqueue_counter==PQUEUE_LENGTH) {
    send_to_char(ch,"Sorry, there are too many requests now, try later.\r\n");
    return;
  }
  program_queue[pqueue_counter]=prg;
  pqueue_counter++;
  if (pqueue_counter==1)	 /* No process is running now so start new process */
    program_fork();
}

/* 
 * swho calls program w. 
 * It shows who is logged on (not in mud, in unix) 
 * and what they are doing
 */
 
ACMD(do_swho) {
  struct program_info swho;
  //  fprintf(stderr,"Entering do_swho");
  swho.args[0]=strdup("w");
  swho.args[1]=NULL;
  swho.input=NULL;
  swho.timeout=3;
  swho.name=strdup("swho");
  add_program(swho, ch); 
}

/* 
 * slast calls program last.
 * It shows last 10 login of specified user or last 10 logins
 * if no user is specified.
 */

ACMD(do_slast) {
  struct program_info slast;
  char arg[128];

  slast.args[0]=strdup("last");
  slast.args[1]=strdup("-10");
  one_argument(argument, arg);
  if (!*arg) 			/* No user specified */
    slast.args[2]=NULL;
  else
    slast.args[2]=strdup(arg);
  slast.args[3]=NULL;
  slast.input=NULL;
  slast.timeout=5;
  slast.name=strdup("slast");
  add_program(slast, ch);
}

/* 
 * grep is normal grep <pattern> <file>
 * Useful is for example:
 * 	grep ../syslog <player> 
 *  or  grep ../log/syslog.2 <player>
 */
 
ACMD(do_grep) {
  struct program_info grep;
  char arg[128];

  argument=one_word(argument, arg);
  skip_spaces(&argument);
  grep.args[0]=strdup("grep");
  /* Ignore cas, becouse one_argument function makes all characters lower. */
  grep.args[1]=strdup("-i");     
  grep.args[2]=strdup(arg);
  grep.args[3]=strdup(argument);
  grep.args[4]=NULL;
  grep.input=NULL;
  grep.timeout=10;
  grep.name=strdup("grep");
  add_program(grep, ch);
}

/*
 * sps calls program ps.
 * I use for example sps aux, sps ux on Linux, but on other
 * platforms it can have another parameters
 */

ACMD(do_sps) {
  struct program_info sps;
  
  skip_spaces(&argument);
  sps.args[0]=strdup("ps");
  if(argument)
    sps.args[1]=strdup(argument);
  else
    sps.args[1]= NULL;
  sps.args[2]= NULL;
  sps.input=NULL;
  sps.timeout=1;
  sps.name=strdup("sps");
  add_program(sps, ch);
}

/*
 * ukill kills runnig child process before his timeout.
 */

ACMD(do_ukill) {
  struct char_data *user;
  char buf[MAX_STRING_LENGTH+1];

  if (pid!=-1) {
    user=get_ch_by_id(program_queue[0].char_id);
    snprintf(buf, sizeof(buf),"&c%s&w has been killed.\r\n", program_queue[0].name);
    kill(pid, 9);	 /* Send SIGKILL to process */
    program_done();
    send_to_char(ch,"Process killed.\r\n");
    if (user) 
      send_to_char(user,buf);
  } else
    send_to_char(ch,"No process is running now.\r\n");  
}

// ZMAIL is found in the arc/utils subdirectory in zmail.c 
// What it does is mail all the data files for a given zone
// to a builder if they have permission to edit the zone.
// 
// I have found that some servers will refuse to accept this
// mail if the source has an invalid return address.  This should
// not be an issue with a production mud.  One way to get around
// the problem is to use a yahoo or hotmail (or pick your favorite
// free mail service) on the receiving end.
//
ACMD(do_zmail) {
  struct program_info zmail;
  char arg[128];

  argument=one_word(argument, arg);
  skip_spaces(&argument);
  if(!arg || !argument || !isdigit(*arg)){
    send_to_char(ch,"Usage: zmail <zone_num> <address>\r\n");
    return;
  }
  //
  // Replace this with whatever mechanism you use to determine whether or
  // not someone may change a zone on your mud.
  //
  //  if(!check_olc_vzone(ch,atoi(arg))){
  int znum = real_zone_by_thing(atoi(arg));
  if(!can_edit_zone(ch,znum)){
    send_to_char(ch,"You do not have permission for that zone.\r\n");
    return;
  }
  send_to_char(ch,"Warning - some addresses refuse this mail, if so, try yahoo or hotmail for service.\r\n");
  zmail.args[0]=strdup("./zmail");
  /*  Ignore cas, becouse one_argument function makes all characters lower.
   */
  zmail.args[3]=NULL;     
  zmail.args[1]=strdup(arg);
  zmail.args[2]=strdup(argument);
  zmail.args[4]=NULL;
  zmail.input=NULL;
  zmail.timeout=6000;
  zmail.name=strdup("zmail");
  add_program(zmail, ch);
}

// Will dump the last 25 lines of the syslog to the player's screen
ACMD(do_catlog) {
  struct program_info tail;
  char arg[128];

  argument=one_argument(argument, arg);
  skip_spaces(&argument);
  tail.args[0]=strdup("tail");
  /*  Ignore cas, becouse one_argument function makes all characters lower.
   */
  tail.args[1]=strdup("-n");
  tail.args[2]=strdup("25");
  tail.args[3]=strdup("../syslog"); 
  tail.args[4]=NULL;
  tail.input=NULL;
  tail.timeout=10;
  tail.name=strdup("tail");
  add_program(tail, ch);
}

// Will dump the entire syslog to the player's screen. (this is a bunch since
// uptimes are on the order of 60-200 days at a time) Prolly best to come up
// with some automatic log rotation scheme, or even better, allow an argument
// to (catlog) or other function that will specify how many hours of log are 
// desired.
ACMD(do_logdump) {
  struct program_info grep;
  char arg[128];

  argument=one_argument(argument, arg);
  skip_spaces(&argument);
  grep.args[0]=strdup("cat");
  /* Ignore cas, becouse one_argument function makes all characters lower.
   */
  grep.args[1]=strdup("../syslog");     
  grep.args[2]=NULL;
  grep.input=NULL;
  grep.timeout=10;
  grep.name=strdup("cat");
  add_program(grep, ch);
}

// Cat the specified file to the users screen
ACMD(do_cat) {
  struct program_info grep;
  char arg[128];

  argument=one_argument(argument, arg);
  skip_spaces(&argument);
  grep.args[0]=strdup("cat");
  /* Ignore cas, becouse one_argument function makes all characters lower.
   */
  grep.args[1]=strdup(arg);     
  grep.args[2]=NULL;
  grep.input=NULL;
  grep.timeout=10;
  grep.name=strdup("cat");
  add_program(grep, ch);
}

// cat the crashlog (syslog.CRASH) to the user's screen
ACMD(do_crashlog) {
  struct program_info grep;
  char arg[128];

  argument=one_argument(argument, arg);
  skip_spaces(&argument);
  grep.args[0]=strdup("cat");
  /* Ignore cas, becouse one_argument function makes all characters lower.
   */
  grep.args[1]=strdup("../syslog.CRASH");     
  grep.args[2]=NULL;
  grep.input=NULL;
  grep.timeout=10;
  grep.name=strdup("cat");
  add_program(grep, ch);
}

// In-game backups - each backup overwrites the last. I should fix this
// to include a date string in the filename.  This will allow us to revert
// to specific "versions" of the server and data files.
ACMD(do_backup){
  struct program_info tar;
  
  //
  // Replace this with whatever mechanism you use on your mud for "trusting"
  // or allowing users to use "system"-like commands.
  // 
  //  if(!can_code(ch)){
  //    send_to_char("You are not authorized for this command.\r\n",ch);
  //    return;
  //  }
  switch(subcmd){
  case SCMD_BACKUP_PLAYERS:
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("ig_player_backup.tgz");
    tar.args[3] = strdup("plralias plrobjs");
    tar.args[4] = strdup("plrvars");
    tar.args[5] = strdup("etc/players");
    tar.input = NULL;
    tar.timeout = 60;
    tar.name=strdup("tar");
    add_program(tar,ch);
    break;
  case SCMD_BACKUP_SOURCE:
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("../src/ig_source_backup.tgz");
    tar.args[3] = strdup("*.c");
    tar.args[4] = strdup("*.h");
    tar.args[5] = strdup("Makefile Makefile.in");
    tar.input = NULL;
    tar.timeout = 60;
    tar.name=strdup("tar");
    add_program(tar,ch);
    break;    
  case SCMD_BACKUP_WORLD:
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("ig_world_backup.tgz");
    tar.args[3] = strdup("world");
    tar.args[4] = NULL;
    tar.input = NULL;
    tar.timeout = 60;
    tar.name = strdup("tar");
    add_program(tar,ch);
    break;
  case SCMD_BACKUP_BACKALL:
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("ig_player_backup.tgz");
    tar.args[3] = strdup("plralias");
    tar.args[4] = strdup("plrobjs plrvars");
    tar.args[5] = strdup("etc/players");
    tar.input = NULL;
    tar.timeout = 60;
    tar.name=strdup("tar");
    add_program(tar,ch);
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("ig_world_backup.tgz");
    tar.args[3] = strdup("world");
    tar.args[4] = NULL;
    tar.input = NULL;
    tar.timeout = 60;
    tar.name = strdup("tar");
    add_program(tar,ch);
    tar.args[0] = strdup("tar");
    tar.args[1] = strdup("zcvf");
    tar.args[2] = strdup("../src/ig_source_backup.tgz");
    tar.args[3] = strdup("*.c");
    tar.args[4] = strdup("*.h");
    tar.args[5] = strdup("Makefile Makefile.in");
    tar.input = NULL;
    tar.timeout = 60;
    tar.name=strdup("tar");
    add_program(tar,ch);
    break;
    
  }
}

/*
 * Adds make functionality
 */
ACMD(do_cmake) {
  struct program_info smake;
  
  //  if (!can_code(ch)) {
  //    send_to_char("Sorry, but you are not cleared for this command.\r\n",ch);
  //    return;
  //  }  
  
  smake.args[0]=strdup("make");
  smake.args[1]=strdup("-C");
  smake.args[2]=strdup("../src"); /* Mud is running in circle/lib directory. 
				     This tells make to work in circle/src. */
  skip_spaces(&argument);
  if (*argument)
    smake.args[3]=strdup(argument);
  else
    smake.args[3]=NULL;
  smake.args[4]=NULL;
  smake.input=NULL;
  smake.timeout=6000;		/* Maybe timeout should be longer? */
  smake.name=strdup("make");
  add_program(smake, ch);
}



                   
